//--------------------------------------------------------------
// File     : stm32_ub_sst25vf016.c
// Datum    : 10.05.2015
// Version  : 1.0
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : STM32F429
// Board    : STM32F429-Board von UB (Version 1.0)
// IDE      : CooCox CoIDE 1.7.4
// GCC      : 4.7 2012q4
// Module   : GPIO, STM32_UB_SPI
// Funktion : SST25VF016 (SPI-Flash 16MBit = 2MByte)
//
// Hinweis  : Das SPI-Flash benutzt die CPU-Pins :
//              PI11 = SPI_CS
//              PF7  = SPI_CLK
//              PF8  = SPI_MISO
//              PF9  = SPI_MOSI
//
// SPI-Settings :  SPI-Mode = SPI_MODE_0_MSB, FRQ-Max = 25MHz (bzw. 80MHz)
//                 SPI5 [CLK=PF7, MOSI=PF9, MISO=PF8]
//                 Chip-Select an PI11
//--------------------------------------------------------------

//--------------------------------------------------------------
// Includes
//--------------------------------------------------------------
#include "stm32_ub_sst25vf016.h"


//--------------------------------------------------------------
// interne Funktionen
//-------------------------------------------------------------- 
void P_SST25VF016_InitIO(void);
void P_SST25VF016_CS(BitAction wert);
ErrorStatus P_SST25VF016_CheckID(void);
void P_SST25VF016_WE(void);
void P_SST25VF016_WD(void);
void P_SST25VF016_WAIT4BUSY(void);
void P_SST25VF016_Delay(volatile uint32_t nCount);


//--------------------------------------------------------------
// Init vom SPI-Flash
// Return_wert :
//  -> ERROR   , wenn SPI-Flash nicht erkannt wurde
//  -> SUCCESS , wenn SPI-Flash gefunden wurde
//--------------------------------------------------------------
ErrorStatus UB_SST25VF016_Init(void)
{
  ErrorStatus ret_wert=ERROR;
  
  // init aller IO-Pins
  P_SST25VF016_InitIO();  
  // init vom SPI-BUS
  ret_wert=UB_SPI5_Init(SPI_MODE_0_MSB);
  if(ret_wert==SUCCESS) { 
    // check der ID vom SPI-Flash
    ret_wert=P_SST25VF016_CheckID();
  }
  
  return(ret_wert);
}


//--------------------------------------------------------------
// Block protection level auslesen
// ret_wert : 0 = kein Schreibschutz
//            1 = obere 64 kByte schreibgeschtzt
//            2 = obere 128 kByte schreibgeschtzt
//            3 = obere 256 kByte schreibgeschtzt
//            4 = obere 512 kByte schreibgeschtzt
//            5 = obere 1 MByte schreibgeschtzt
//            6 = komplette 2 MByte schreibgeschtzt
//            7 = komplette 2 MByte schreibgeschtzt (default nach PowerOn)
//--------------------------------------------------------------
uint8_t UB_SST25VF016_GetProtectionLevel(void)
{
  uint8_t status_reg,ret_wert=0;

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // Read-Status-Register CMD senden
  UB_SPI5_SendByte(SST25VF016_RSR_CMD);

  // StatusRegister auslesen (1 Byte)
  status_reg=UB_SPI5_SendByte(0x55);

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);

  // status register auswerten
  if((status_reg&SST25VF016_BP0_BIT)!=0) ret_wert|=0x01;
  if((status_reg&SST25VF016_BP1_BIT)!=0) ret_wert|=0x02;
  if((status_reg&SST25VF016_BP2_BIT)!=0) ret_wert|=0x04;

  return(ret_wert);
}

//--------------------------------------------------------------
// Block protection level einstellen
// mode : 0 = kein Schreibschutz
//        1 = obere 64 kByte schreibgeschtzt
//        2 = obere 128 kByte schreibgeschtzt
//        3 = obere 256 kByte schreibgeschtzt
//        4 = obere 512 kByte schreibgeschtzt
//        5 = obere 1 MByte schreibgeschtzt
//        6 = komplette 2 MByte schreibgeschtzt
//        7 = komplette 2 MByte schreibgeschtzt (default nach PowerOn)
//--------------------------------------------------------------
void UB_SST25VF016_SetProtectionLevel(uint8_t mode)
{
  uint8_t status_reg=0;

  if(mode>7) mode=7;

  // status register setzen
  if((mode&0x01)!=0)  status_reg|=SST25VF016_BP0_BIT;
  if((mode&0x02)!=0)  status_reg|=SST25VF016_BP1_BIT;
  if((mode&0x04)!=0)  status_reg|=SST25VF016_BP2_BIT;

  // write enable
  P_SST25VF016_WE();

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // WriteStatusRegsiter CMD senden
  UB_SPI5_SendByte(SST25VF016_WSR_CMD);

  // StatusRegister schreiben
  UB_SPI5_SendByte(status_reg);

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);

}


//--------------------------------------------------------------
// das komplette SPI-Flash loeschen
// (hebt Schreibschutz kurzzeitig auf)
//--------------------------------------------------------------
void UB_SST25VF016_EraseFlash(void)
{
  uint8_t protection;

  // aktuellen Schreibschutz auslesen
  protection=UB_SST25VF016_GetProtectionLevel();

  // schreibschutz aufheben
  UB_SST25VF016_SetProtectionLevel(0);

  // write enable
  P_SST25VF016_WE();

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET); 

  // ChipErase CMD senden
  UB_SPI5_SendByte(SST25VF016_CLRALL_CMD); 

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);

  // warten bis erase fertig
  P_SST25VF016_WAIT4BUSY();

  // protection level wieder herstellen
  UB_SST25VF016_SetProtectionLevel(protection);
}


//--------------------------------------------------------------
// ein Byte vom SPI-Flash von adr auslesen  (mit max 25MHz)
// adr : [0...1FFFFF]
// ret_wert : [0...255]
//--------------------------------------------------------------
uint8_t UB_SST25VF016_ReadByte(uint32_t adr)
{
  uint8_t ret_wert=0;
  uint8_t adr_b2,adr_b1,adr_b0;

  if(adr>=SST25VF016_SIZE) return 0;

  // adresse umwandeln
  adr_b0=(adr&0xFF);
  adr_b1=((adr>>8)&0xFF);
  adr_b2=((adr>>16)&0xFF);

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // Read-Byte CMD senden
  UB_SPI5_SendByte(SST25VF016_RB_CMD);

  // Adresse senden (3 Bytes)
  UB_SPI5_SendByte(adr_b2);
  UB_SPI5_SendByte(adr_b1);
  UB_SPI5_SendByte(adr_b0);

  // Daten auslesen (1 Byte)
  ret_wert=UB_SPI5_SendByte(0x55);

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);

  return(ret_wert);
}


//--------------------------------------------------------------
// ein Byte ins SPI-Flash an adr schreiben
// byte muss mit 0xFF gelscht sein
// und der Block darf nicht schreibgeschuetzt sein
//
// adr : [0...1FFFFF]
// value : [0...255]
//--------------------------------------------------------------
void UB_SST25VF016_WriteByte(uint32_t adr, uint8_t value)
{
  uint8_t adr_b2,adr_b1,adr_b0;

  if(adr>=SST25VF016_SIZE) return;

  // adresse umwandeln
  adr_b0=(adr&0xFF);
  adr_b1=((adr>>8)&0xFF);
  adr_b2=((adr>>16)&0xFF);

  // write enable
  P_SST25VF016_WE();

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // Write-Byte CMD senden
  UB_SPI5_SendByte(SST25VF016_WB_CMD);

  // Adresse senden (3 Bytes)
  UB_SPI5_SendByte(adr_b2);
  UB_SPI5_SendByte(adr_b1);
  UB_SPI5_SendByte(adr_b0);

  // Daten schreiben (1 Byte)
  UB_SPI5_SendByte(value);

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);

  // warten bis write fertig
  P_SST25VF016_WAIT4BUSY();
}


//--------------------------------------------------------------
// liet einen Block von Bytes aus dem externen SPI-Flash
// ab einer Adresse (mit max 25MHz)
//
// ptrBuffer : Pointer auf den Start vom Block der gelesen wird
// startAdr  : Start Adresse vom Flash ab der gelesen wird
// lenBuffer  : [1 bis SST25VF016_SIZE]
//--------------------------------------------------------------
void UB_SST25VF016_ReadBlock(uint8_t* ptrBuffer, uint32_t startAdr, uint32_t lenBuffer)
{
  uint8_t adr_b2,adr_b1,adr_b0;

  if(startAdr>=SST25VF016_SIZE) return;
  if(lenBuffer==0) return;
  if(startAdr+lenBuffer>SST25VF016_SIZE) lenBuffer=SST25VF016_SIZE-startAdr;

  // adresse umwandeln
  adr_b0=(startAdr&0xFF);
  adr_b1=((startAdr>>8)&0xFF);
  adr_b2=((startAdr>>16)&0xFF);

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // Read-Byte CMD senden
  UB_SPI5_SendByte(SST25VF016_RB_CMD);

  // Adresse senden (3 Bytes)
  UB_SPI5_SendByte(adr_b2);
  UB_SPI5_SendByte(adr_b1);
  UB_SPI5_SendByte(adr_b0);

  // alle Daten vom Flash lesen
  for(; lenBuffer != 0x00; lenBuffer--) {
	// Byte auslesen
    *ptrBuffer++ = UB_SPI5_SendByte(0x55);
  }

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);
}


//--------------------------------------------------------------
// schreibt einen Block von Bytes in das externen SPI-Flash
// ab einer Adresse
// Daten muessen mit 0xFF geloescht sein
// und der Block darf nicht schreibgeschuetzt sein
//
// ptrBuffer : Pointer auf den Start vom Block der geschrieben wird
// startAdr  : Start Adresse vom Flash ab der geschrieben wird
// lenBuffer : anzahl der Daten die geschrieben werden sollen
//             [2 bis SST25VF016_SIZE]
//--------------------------------------------------------------
void UB_SST25VF016_WriteBlock(uint8_t* ptrBuffer, uint32_t startAdr, uint32_t lenBuffer)
{
  uint32_t n;
  uint8_t adr_b2,adr_b1,adr_b0,wert,rest;

  if(startAdr>=SST25VF016_SIZE) return;
  if(lenBuffer<2) return;
  if(startAdr+lenBuffer>SST25VF016_SIZE) lenBuffer=SST25VF016_SIZE-startAdr;

  // adresse umwandeln
  adr_b0=(startAdr&0xFF);
  adr_b1=((startAdr>>8)&0xFF);
  adr_b2=((startAdr>>16)&0xFF);

  // check ob gerade anzahl von bytes
  rest=0;
  if((lenBuffer&0x01)!=0) {
    lenBuffer--; // ein byte weniger schreiben
    rest=1;
  }

  // write enable
  P_SST25VF016_WE();

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // WriteWord CMD senden
  UB_SPI5_SendByte(SST25VF016_WW_CMD);

  // Adresse senden (3 Bytes)
  UB_SPI5_SendByte(adr_b2);
  UB_SPI5_SendByte(adr_b1);
  UB_SPI5_SendByte(adr_b0);

  // alle Daten ins Flash schreiben (immer zwei Bytes)
  for(n=0;n<lenBuffer;n+=2) {
    // byte 1 schreiben
    wert=*ptrBuffer++;
    UB_SPI5_SendByte(wert);
    // byte 2 schreiben
    wert=*ptrBuffer++;
    UB_SPI5_SendByte(wert);
    // ChipSelect auf Hi
    P_SST25VF016_CS(Bit_SET);
    // warten bis write fertig
    P_SST25VF016_WAIT4BUSY();
    if((n+2)<lenBuffer) {
      // ChipSelect auf Lo
      P_SST25VF016_CS(Bit_RESET);
      // WriteWord CMD nochmal senden
      UB_SPI5_SendByte(SST25VF016_WW_CMD);
    }
    startAdr+=2;
  }

  // write disable
  P_SST25VF016_WD();

  // event das letzte Byte einzeln schreiben
  if(rest==1) {
    wert=*ptrBuffer++;
    UB_SST25VF016_WriteByte(startAdr,wert);
  }
}


//--------------------------------------------------------------
// interne Funktion
// Init aller IO-Pins fuer den SST25VF016
//--------------------------------------------------------------
void P_SST25VF016_InitIO(void)
{	
  GPIO_InitTypeDef       GPIO_InitStructure;

  // ChipSelect-Pin
  RCC_AHB1PeriphClockCmd(SST25VF016_CS_GPIO_CLK, ENABLE);
	  
  GPIO_InitStructure.GPIO_Pin = SST25VF016_CS_PIN;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_Init(SST25VF016_CS_GPIO_PORT, &GPIO_InitStructure);	

  P_SST25VF016_CS(Bit_SET);
}

//--------------------------------------------------------------
// interne Funktion
// ChipSelect-Pin schalten
// wert [Bit_SET, Bit_RESET]
//--------------------------------------------------------------
void P_SST25VF016_CS(BitAction wert)
{
  if (wert == Bit_RESET) {
    SST25VF016_CS_GPIO_PORT->BSRRH = SST25VF016_CS_PIN;
  }
  else {
    SST25VF016_CS_GPIO_PORT->BSRRL = SST25VF016_CS_PIN;
  }
}

//--------------------------------------------------------------
// interne Funktion
// check der ID vom SPI-Flash
//--------------------------------------------------------------
ErrorStatus P_SST25VF016_CheckID(void)
{
  ErrorStatus ret_wert=SUCCESS;
  uint8_t manufacture_id=0, device_id=0;

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // Read-ID CMD senden
  UB_SPI5_SendByte(SST25VF016_RID_CMD);
  
  // Adresse 0x000000 senden (3 Bytes)
  UB_SPI5_SendByte(0x00);
  UB_SPI5_SendByte(0x00);
  UB_SPI5_SendByte(0x00);
  
  // ID auslesen (2 Bytes)
  manufacture_id=UB_SPI5_SendByte(0x55);
  device_id=UB_SPI5_SendByte(0x55);
 
  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);  

  // id auswerten
  if(manufacture_id!=SST25VF016_MAN_ID) ret_wert=ERROR;
  if(device_id!=SST25VF016_DEV_ID) ret_wert=ERROR;

  return(ret_wert);
}

//--------------------------------------------------------------
// interne Funktion
// WriteEnable
//--------------------------------------------------------------
void P_SST25VF016_WE(void)
{
  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // WriteEnable CMD senden
  UB_SPI5_SendByte(SST25VF016_WE_CMD);

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);
}

//--------------------------------------------------------------
// interne Funktion
// WriteDisable
//--------------------------------------------------------------
void P_SST25VF016_WD(void)
{
  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // WriteDisable CMD senden
  UB_SPI5_SendByte(SST25VF016_WD_CMD);

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);
}

//--------------------------------------------------------------
// interne Funktion
// status Register vom SPI-Flash auslesen
// und warten bis Busy-Bit disable
//--------------------------------------------------------------
void P_SST25VF016_WAIT4BUSY(void)
{
  uint8_t ok,status_reg;

  // ChipSelect auf Lo
  P_SST25VF016_CS(Bit_RESET);

  // Read-Status-Register CMD senden
  UB_SPI5_SendByte(SST25VF016_RSR_CMD);

  ok=0;
  // warte bis Busy=0
  do {
    // Register auslesen (1 Byte)
    status_reg=UB_SPI5_SendByte(0x55);
    // busy auswerten
    if((status_reg&SST25VF016_BUSY_BIT)==0) ok=1;
  }while(ok==0);

  // ChipSelect auf Hi
  P_SST25VF016_CS(Bit_SET);  
}

//--------------------------------------------------------------
// interne Funktion
// kleine Pause
//-------------------------------------------------------------- 
void P_SST25VF016_Delay(volatile uint32_t nCount)
{
  volatile uint32_t index = 0;

  for(index = nCount; index != 0; index--);
}
